@@ -1,12 +1,13 @@
 #[cfg(test)]
 mod tests {
-    use crate::loan::{make_keypair, Borrower0, Lender0};
+    use std::time::SystemTime;
+
+    use crate::loan::{oracle, Borrower0, Lender0};
     use anyhow::{Context, Result};
     use bitcoin_hashes::Hash;
     use elements::bitcoin::util::psbt::serialize::Serialize;
-    use elements::bitcoin::{Amount, PublicKey};
+    use elements::bitcoin::{Amount, Network, PrivateKey, PublicKey};
     use elements::script::Builder;
-    use elements::secp256k1_zkp::rand::thread_rng;
     use elements::secp256k1_zkp::{SecretKey, SECP256K1};
     use elements::sighash::SigHashCache;
     use elements::{
@@ -15,12 +16,16 @@ mod tests {
     };
     use elements_harness::Elementsd;
     use elements_rpc::ElementsRpc;
+    use rand::{CryptoRng, RngCore, SeedableRng};
+    use rand_chacha::ChaChaRng;
     use testcontainers::clients::Cli;
 
     #[tokio::test]
     async fn borrow_and_repay() {
         init_logger();
 
+        let mut rng = ChaChaRng::seed_from_u64(0);
+
         let tc_client = Cli::default();
         let (client, _container) = {
             let blockchain = Elementsd::new(&tc_client, "0.18.1.9").unwrap();
@@ -33,6 +38,7 @@ mod tests {
 
         let bitcoin_asset_id = client.get_bitcoin_asset_id().await.unwrap();
         let usdt_asset_id = client.issueasset(40.0, 0.0, false).await.unwrap().asset;
+        let (_oracle_sk, oracle_pk) = make_keypair(&mut rng);
 
         let miner_address = client.get_new_segwit_confidential_address().await.unwrap();
 
@@ -43,7 +49,7 @@ mod tests {
         client.generatetoaddress(10, &miner_address).await.unwrap();
 
         let (borrower, borrower_wallet) = {
-            let mut wallet = Wallet::new();
+            let mut wallet = Wallet::new(&mut rng);
 
             let collateral_amount = Amount::ONE_BTC;
 
@@ -74,7 +80,7 @@ mod tests {
             let timelock = 10;
 
             let borrower = Borrower0::new(
-                &mut thread_rng(),
+                &mut rng,
                 {
                     let wallet = wallet.clone();
                     |amount, asset| async move { wallet.find_inputs(asset, amount).await }
@@ -95,12 +101,15 @@ mod tests {
 
         let (lender, _lender_address) = {
             let address = client.get_new_segwit_confidential_address().await.unwrap();
+            let address_blinder = blinding_key(&client, &address).await.unwrap();
 
             let lender = Lender0::new(
-                &mut thread_rng(),
+                &mut rng,
                 bitcoin_asset_id,
                 usdt_asset_id,
                 address.clone(),
+                address_blinder,
+                oracle_pk,
             )
             .unwrap();
 
@@ -111,8 +120,8 @@ mod tests {
 
         let lender = lender
             .interpret(
-                &mut thread_rng(),
-                &SECP256K1,
+                &mut rng,
+                SECP256K1,
                 {
                     let client = client.clone();
                     |amount, asset| async move { find_inputs(&client, asset, amount).await }
@@ -124,7 +133,7 @@ mod tests {
             .unwrap();
         let loan_response = lender.loan_response();
 
-        let borrower = borrower.interpret(&SECP256K1, loan_response).unwrap();
+        let borrower = borrower.interpret(SECP256K1, loan_response).unwrap();
         let loan_transaction = borrower
             .sign({
                 let wallet = borrower_wallet.clone();
@@ -150,8 +159,8 @@ mod tests {
 
         let loan_repayment_transaction = borrower
             .loan_repayment_transaction(
-                &mut thread_rng(),
-                &SECP256K1,
+                &mut rng,
+                SECP256K1,
                 {
                     let borrower_wallet = borrower_wallet.clone();
                     |amount, asset| async move { borrower_wallet.find_inputs(asset, amount).await }
@@ -165,18 +174,15 @@ mod tests {
         client
             .send_raw_transaction(&loan_repayment_transaction)
             .await
-            .unwrap();
-    }
-
-    fn init_logger() {
-        // force enabling log output
-        let _ = env_logger::builder().is_test(true).try_init();
+            .expect("could not repay loan to reclaim collateral");
     }
 
     #[tokio::test]
     async fn lend_and_liquidate() {
         init_logger();
 
+        let mut rng = ChaChaRng::seed_from_u64(0);
+
         let tc_client = Cli::default();
         let (client, _container) = {
             let blockchain = Elementsd::new(&tc_client, "0.18.1.9").unwrap();
@@ -189,6 +195,7 @@ mod tests {
 
         let bitcoin_asset_id = client.get_bitcoin_asset_id().await.unwrap();
         let usdt_asset_id = client.issueasset(40.0, 0.0, false).await.unwrap().asset;
+        let (_oracle_sk, oracle_pk) = make_keypair(&mut rng);
 
         let miner_address = client.get_new_segwit_confidential_address().await.unwrap();
         client
@@ -198,7 +205,7 @@ mod tests {
         client.generatetoaddress(10, &miner_address).await.unwrap();
 
         let (borrower, borrower_wallet) = {
-            let mut wallet = Wallet::new();
+            let mut wallet = Wallet::new(&mut rng);
 
             let collateral_amount = Amount::ONE_BTC;
 
@@ -226,10 +233,10 @@ mod tests {
 
             client.generatetoaddress(1, &miner_address).await.unwrap();
 
-            let timelock = 0;
+            let timelock = client.get_blockcount().await.unwrap() + 5;
 
             let borrower = Borrower0::new(
-                &mut thread_rng(),
+                &mut rng,
                 {
                     let wallet = wallet.clone();
                     |amount, asset| async move { wallet.find_inputs(asset, amount).await }
@@ -250,12 +257,15 @@ mod tests {
 
         let (lender, _lender_address) = {
             let address = client.get_new_segwit_confidential_address().await.unwrap();
+            let address_blinder = blinding_key(&client, &address).await.unwrap();
 
             let lender = Lender0::new(
-                &mut thread_rng(),
+                &mut rng,
                 bitcoin_asset_id,
                 usdt_asset_id,
                 address.clone(),
+                address_blinder,
+                oracle_pk,
             )
             .unwrap();
 
@@ -266,8 +276,8 @@ mod tests {
 
         let lender = lender
             .interpret(
-                &mut thread_rng(),
-                &SECP256K1,
+                &mut rng,
+                SECP256K1,
                 {
                     let client = client.clone();
                     |amount, asset| async move { find_inputs(&client, asset, amount).await }
@@ -279,7 +289,7 @@ mod tests {
             .unwrap();
         let loan_response = lender.loan_response();
 
-        let borrower = borrower.interpret(&SECP256K1, loan_response).unwrap();
+        let borrower = borrower.interpret(SECP256K1, loan_response).unwrap();
         let loan_transaction = borrower
             .sign(|transaction| async move { Ok(borrower_wallet.sign_all_inputs(transaction)) })
             .await
@@ -298,16 +308,295 @@ mod tests {
             .await
             .unwrap();
 
-        client.generatetoaddress(1, &miner_address).await.unwrap();
-
         let liquidation_transaction = lender
-            .liquidation_transaction(&mut thread_rng(), &SECP256K1, Amount::from_sat(1))
+            .liquidation_transaction(&mut rng, SECP256K1, Amount::from_sat(1))
+            .await
             .unwrap();
 
         client
             .send_raw_transaction(&liquidation_transaction)
             .await
+            .expect_err("could liquidate before loan term");
+
+        client.generatetoaddress(5, &miner_address).await.unwrap();
+
+        client
+            .send_raw_transaction(&liquidation_transaction)
+            .await
+            .expect("could not liquidate after loan term");
+    }
+
+    #[tokio::test]
+    async fn lend_and_dynamic_liquidate() {
+        init_logger();
+
+        let mut rng = ChaChaRng::seed_from_u64(0);
+
+        let tc_client = Cli::default();
+        let (client, _container) = {
+            let blockchain = Elementsd::new(&tc_client, "0.18.1.9").unwrap();
+
+            (
+                elements_rpc::Client::new(blockchain.node_url.clone().into()).unwrap(),
+                blockchain,
+            )
+        };
+
+        let bitcoin_asset_id = client.get_bitcoin_asset_id().await.unwrap();
+        let usdt_asset_id = client.issueasset(40.0, 0.0, false).await.unwrap().asset;
+        let (oracle_sk, oracle_pk) = make_keypair(&mut rng);
+
+        let miner_address = client.get_new_segwit_confidential_address().await.unwrap();
+        client
+            .send_asset_to_address(&miner_address, Amount::from_btc(5.0).unwrap(), None)
+            .await
             .unwrap();
+        client.generatetoaddress(10, &miner_address).await.unwrap();
+
+        let (borrower, borrower_wallet) = {
+            let mut wallet = Wallet::new(&mut rng);
+
+            let collateral_amount = Amount::ONE_BTC;
+
+            let address = wallet.address();
+            let address_blinding_sk = wallet.dump_blinding_sk();
+
+            // fund borrower address with bitcoin
+            let txid = client
+                .send_asset_to_address(&address, collateral_amount * 2, Some(bitcoin_asset_id))
+                .await
+                .unwrap();
+
+            wallet.add_known_utxo(&client, txid).await;
+
+            // fund wallet with some usdt to pay back the loan later on
+            let txid = client
+                .send_asset_to_address(
+                    &address,
+                    Amount::from_btc(2.0).unwrap(),
+                    Some(usdt_asset_id),
+                )
+                .await
+                .unwrap();
+            wallet.add_known_utxo(&client, txid).await;
+
+            client.generatetoaddress(1, &miner_address).await.unwrap();
+
+            let timelock = client.get_blockcount().await.unwrap() + 100;
+
+            let borrower = Borrower0::new(
+                &mut rng,
+                {
+                    let wallet = wallet.clone();
+                    |amount, asset| async move { wallet.find_inputs(asset, amount).await }
+                },
+                address.clone(),
+                address_blinding_sk,
+                collateral_amount,
+                Amount::ONE_SAT,
+                timelock,
+                bitcoin_asset_id,
+                usdt_asset_id,
+            )
+            .await
+            .unwrap();
+
+            (borrower, wallet)
+        };
+
+        let (lender, _lender_address) = {
+            let address = client.get_new_segwit_confidential_address().await.unwrap();
+            let address_blinder = blinding_key(&client, &address).await.unwrap();
+
+            let lender = Lender0::new(
+                &mut rng,
+                bitcoin_asset_id,
+                usdt_asset_id,
+                address.clone(),
+                address_blinder,
+                oracle_pk,
+            )
+            .unwrap();
+
+            (lender, address)
+        };
+
+        let loan_request = borrower.loan_request();
+
+        let lender = lender
+            .interpret(
+                &mut rng,
+                SECP256K1,
+                {
+                    let client = client.clone();
+                    |amount, asset| async move { find_inputs(&client, asset, amount).await }
+                },
+                loan_request,
+                38_000, // value of 1 BTC as of 18.06.2021
+            )
+            .await
+            .unwrap();
+        let loan_response = lender.loan_response();
+
+        let borrower = borrower.interpret(SECP256K1, loan_response).unwrap();
+        let loan_transaction = borrower
+            .sign(|transaction| async move { Ok(borrower_wallet.sign_all_inputs(transaction)) })
+            .await
+            .unwrap();
+
+        let loan_transaction = lender
+            .finalise_loan(loan_transaction, {
+                let client = client.clone();
+                |transaction| async move { client.sign_raw_transaction(&transaction).await }
+            })
+            .await
+            .unwrap();
+
+        client
+            .send_raw_transaction(&loan_transaction)
+            .await
+            .unwrap();
+
+        // Oracle message too early:
+        {
+            // before contract creation
+            let now = SystemTime::now()
+                .duration_since(SystemTime::UNIX_EPOCH)
+                .unwrap()
+                .as_secs();
+            let current_timestamp = now - 3600;
+            // price dips below minimum
+            let current_btc_price = 0;
+
+            let oracle_msg = oracle::Message::new(current_btc_price, current_timestamp);
+            let oracle_sig = SECP256K1.sign(&oracle_msg.message_hash(), &oracle_sk);
+
+            let liquidation_transaction = lender
+                .dynamic_liquidation_transaction(
+                    &mut rng,
+                    SECP256K1,
+                    oracle_msg,
+                    oracle_sig,
+                    Amount::ONE_SAT,
+                )
+                .await
+                .unwrap();
+
+            client
+                .send_raw_transaction(&liquidation_transaction)
+                .await
+                .expect_err("could liquidate with proof of dip before contract creation");
+        }
+
+        // Price too high:
+        {
+            // fast forward
+            let now = SystemTime::now()
+                .duration_since(SystemTime::UNIX_EPOCH)
+                .unwrap()
+                .as_secs();
+            let current_timestamp = now + 3600;
+            // price remains way above threshold
+            let current_btc_price = 1_000_000;
+
+            let oracle_msg = oracle::Message::new(current_btc_price, current_timestamp);
+            let oracle_sig = SECP256K1.sign(&oracle_msg.message_hash(), &oracle_sk);
+
+            let liquidation_transaction = lender
+                .dynamic_liquidation_transaction(
+                    &mut rng,
+                    SECP256K1,
+                    oracle_msg,
+                    oracle_sig,
+                    Amount::ONE_SAT,
+                )
+                .await
+                .unwrap();
+
+            client
+                .send_raw_transaction(&liquidation_transaction)
+                .await
+                .expect_err("could liquidate with proof of dip above threshold");
+        }
+
+        // Not signed by oracle:
+        {
+            // fast forward
+            let now = SystemTime::now()
+                .duration_since(SystemTime::UNIX_EPOCH)
+                .unwrap()
+                .as_secs();
+            let current_timestamp = now + 3600;
+            // price dips below minimum
+            let current_btc_price = 0;
+
+            let oracle_msg = oracle::Message::new(current_btc_price, current_timestamp);
+
+            // signed by a fake oracle
+            let (fake_oracle_sk, _) = make_keypair(&mut rng);
+            let oracle_sig = SECP256K1.sign(&oracle_msg.message_hash(), &fake_oracle_sk);
+
+            let liquidation_transaction = lender
+                .dynamic_liquidation_transaction(
+                    &mut rng,
+                    SECP256K1,
+                    oracle_msg,
+                    oracle_sig,
+                    Amount::ONE_SAT,
+                )
+                .await
+                .unwrap();
+
+            client
+                .send_raw_transaction(&liquidation_transaction)
+                .await
+                .expect_err("could liquidate with invalid valid proof of dip");
+        }
+
+        // Success:
+        {
+            // fast forward
+            let now = SystemTime::now()
+                .duration_since(SystemTime::UNIX_EPOCH)
+                .unwrap()
+                .as_secs();
+            let current_timestamp = now + 3600;
+            // price dips below minimum
+            let current_btc_price = 0;
+
+            let oracle_msg = oracle::Message::new(current_btc_price, current_timestamp);
+            let oracle_sig = SECP256K1.sign(&oracle_msg.message_hash(), &oracle_sk);
+
+            let liquidation_transaction = lender
+                .dynamic_liquidation_transaction(
+                    &mut rng,
+                    SECP256K1,
+                    oracle_msg,
+                    oracle_sig,
+                    Amount::ONE_SAT,
+                )
+                .await
+                .unwrap();
+
+            client
+                .send_raw_transaction(&liquidation_transaction)
+                .await
+                .expect("could not liquidate with valid proof of dip");
+        }
+    }
+
+    fn init_logger() {
+        // force enabling log output
+        let _ = env_logger::builder().is_test(true).try_init();
+    }
+
+    async fn blinding_key(client: &elements_rpc::Client, address: &Address) -> Result<SecretKey> {
+        let master_blinding_key = client.dumpmasterblindingkey().await?;
+        let master_blinding_key = hex::decode(master_blinding_key)?;
+
+        let sk = derive_blinding_key(master_blinding_key, address.script_pubkey())?;
+
+        Ok(sk)
     }
 
     async fn find_inputs(
@@ -363,9 +652,12 @@ mod tests {
     }
 
     impl Wallet {
-        pub fn new() -> Self {
-            let (sk, pk) = make_keypair(&mut thread_rng());
-            let (blinder_sk, blinder_pk) = make_keypair(&mut thread_rng());
+        pub fn new<R>(rng: &mut R) -> Self
+        where
+            R: RngCore + CryptoRng,
+        {
+            let (sk, pk) = make_keypair(rng);
+            let (blinder_sk, blinder_pk) = make_keypair(rng);
 
             let address = Address::p2wpkh(&pk, Some(blinder_pk.key), &AddressParams::ELEMENTS);
 
@@ -490,4 +782,21 @@ mod tests {
             tx_to_sign
         }
     }
+
+    fn make_keypair<R>(rng: &mut R) -> (SecretKey, PublicKey)
+    where
+        R: RngCore + CryptoRng,
+    {
+        let sk = SecretKey::new(rng);
+        let pk = PublicKey::from_private_key(
+            SECP256K1,
+            &PrivateKey {
+                compressed: true,
+                network: Network::Regtest,
+                key: sk,
+            },
+        );
+
+        (sk, pk)
+    }
 }
